(ns ^{:doc "Application Description Language: generate swagger routes."
      :author "Simon Brooke"}
  adl.to-swagger
  (:require [adl-support.utils :refer :all]
            [adl.to-hugsql-queries :refer [generate-documentation queries]]
            [clj-time.core :as t]
            [clj-time.format :as f]
            [clojure.java.io :refer [file make-parents writer]]
            [clojure.pprint :refer [pprint]]
            [clojure.string :as s]
            [clojure.xml :as x]))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;
;;;; adl.to-swagger: generate swagger routes.
;;;;
;;;; This program is free software; you can redistribute it and/or
;;;; modify it under the terms of the GNU General Public License
;;;; as published by the Free Software Foundation; either version 2
;;;; of the License, or (at your option) any later version.
;;;;
;;;; This program is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.
;;;;
;;;; You should have received a copy of the GNU General Public License
;;;; along with this program; if not, write to the Free Software
;;;; Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
;;;; USA.
;;;;
;;;; Copyright (C) 2018 Simon Brooke
;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; The overall structure of this has quite closely to follow the structure of
;;; to-hugsql-queries, because essentially we need one JSON entry point to wrap
;;; each query.

(defn file-header
  "TODO: Nothing here works yet."
  [application]
  (list
    'ns
    (symbol (str (safe-name (:name (:attrs application))) ".routes.auto-api"))
    (str "Swagger routes for " (:name (:attrs application))
         " auto-generated by [Application Description Language framework](https://github.com/simon-brooke/adl) at "
         (f/unparse (f/formatters :basic-date-time) (t/now)))
    (list
      :require
      '[reitit.swagger :as swagger]
      '[reitit.swagger-ui :as swagger-ui]
      '[reitit.ring.coercion :as coercion]
      '[reitit.coercion.spec :as spec-coercion]
      '[reitit.ring.middleware.muuntaja :as muuntaja]
      '[reitit.ring.middleware.multipart :as multipart]
      '[reitit.ring.middleware.parameters :as parameters]
      '[placenames.middleware.formats :as formats]
      '[placenames.middleware.exception :as exception]
      '[placenames.routes.auto-jason :as aj]
      '[ring.util.http-response :refer :all]
      '[clojure.java.io :as io])))



(defn def-routes
  "Generate Swagger routes for all queries implied by this ADL `application` spec."
  ;; THIS ISN'T NEARLY FINISHED!
  ([application]
   (list 'defn 'auto-api-routes []
  ["/api"
   {:coercion spec-coercion/coercion
    :muuntaja formats/instance
    :swagger {:id ::api}
    :middleware [;; query-params & form-params
                 parameters/parameters-middleware
                 ;; content-negotiation
                 muuntaja/format-negotiate-middleware
                 ;; encoding response body
                 muuntaja/format-response-middleware
                 ;; exception handling
                 exception/exception-middleware
                 ;; decoding request body
                 muuntaja/format-request-middleware
                 ;; coercing response bodys
                 coercion/coerce-response-middleware
                 ;; coercing request parameters
                 coercion/coerce-request-middleware
                 ;; multipart
                 multipart/multipart-middleware]}]
     (map #(def-routes application %)
          (children-with-tag application :entity)))
  ([application entity]
   [(str "/" (safe-name entity))
    {:get (make-get-route entity)
  (cons
    'defroutes
    (cons
      'auto-rest-routes
      (map
        #(let [handler (handlers-map %)]
           (list
             (symbol (s/upper-case (name (:method handler))))
             (str "/json/auto/" (safe-name (:name handler)))
             'request
             (list
               'route/restricted
               (list (:name handler) 'request))))
        (sort
          (keys handlers-map)))))}])))


(defn to-swagger
  "Generate a Swagger API for all queries implied by this ADL `application` spec."
  [application]
  (let [filepath (str
                   *output-path*
                   "src/"
                   (safe-name (:name (:attrs application)))
                   "/routes/auto_api.clj")]
    (make-parents filepath)
    (do-or-warn
      (do
        (spit
          filepath
          (s/join
            "\n\n"
            (cons
              (file-header application)
              (map
                (fn [q]
                  (str
                    ;; THIS ISN'T NEARLY FINISHED!
                    ))
                (sort
                  #(compare (:name %1) (:name %2))
                  (vals
                    (queries application)))))))
        (if (pos? *verbosity*)
          (*warn* (str "\tGenerated " filepath)))))))
